from .pkg_resources.lipd.LiPD_Library import *
from .pkg_resources.timeseries.Convert import *
from .pkg_resources.timeseries.TimeSeries_Library import *
from .pkg_resources.doi.doi_main import doi
from .pkg_resources.excel.excel_main import excel
from .pkg_resources.noaa.noaa_main import noaa


def setDir():
    """
    Set the current working directory by providing a directory path.
    (ex. /Path/to/files)
    :param path: (str) Directory path
    """
    global path
    path = set_source()
    lipd_lib.setDir(path)
    return path


def loadLipd(filename):
    """
    Load a single LiPD file into the workspace. File must be located in the current working directory.
    (ex. loadLiPD NAm-ak000.lpd)
    :param filename: (str) LiPD filename
    """
    lipd_lib.loadLipd(filename)
    print("Process Complete")
    return


def loadLipds():
    """
    Load all LiPD files in the current working directory into the workspace.
    """
    lipd_lib.loadLipds()
    print("Process Complete")
    return


# ANALYSIS - LIPD


def showCsv(filename):
    """
    Show CSV data for one LiPD
    :param filename:
    :return:
    """
    lipd_lib.showCsv(filename)
    print("Process Complete")
    return

def showMetadata(filename):
    """
    Display the contents of the specified LiPD file. (Must be previously loaded into the workspace)
    (ex. displayLiPD NAm-ak000.lpd)
    :param filename: (str) LiPD filename
    """
    lipd_lib.showMetadata(filename)
    print("Process Complete")
    return

def showLipds():
    """
    Prints the names of all LiPD files in the LiPD_Library
    """
    lipd_lib.showLipds()
    print("Process Complete")
    return


def showMap(filename):
    """

    :param filename:
    :return:
    """
    lipd_lib.showMap(filename)
    print("Process Complete")
    return


# ANALYSIS - TIME SERIES


def extractTimeSeries():
    """
    Create a TimeSeries using the current files in LiPD_Library.
    :return: (obj) TimeSeries_Library
    """
    # Loop over the LiPD objects in the LiPD_Library
    for k, v in lipd_lib.get_master().items():
        # Get metadata from this LiPD object. Convert it. Pass TSO metadata to the TS_Library.
        ts_lib.loadTsos(v.get_name_ext(), convert.ts_extract_main(v.get_master()))

    print("Process Complete")
    return


def exportTimeSeries():
    """
    Export TimeSeries back to LiPD Library. Updates information in LiPD objects.
    """
    l = []
    # Get all TSOs from TS_Library, and add them to a list
    for k, v in ts_lib.get_master().items():
        l.append({'name': v.get_lpd_name(), 'data': v.get_master()})
    # Send the TSOs list through to be converted. Then let the LiPD_Library load the metadata into itself.
    lipd_lib.load_tsos(convert.lipd_extract_main(l))
    print("Process Complete")
    return


def showTso(name):
    """
    Show contents of one TimeSeries object.
    :param name:
    :return:
    """
    ts_lib.showTso(name)
    print("Process Complete")
    return


def showTsos():
    """
    Prints the names of all TimeSeries objects in the TimeSeries_Library
    :return:
    """
    ts_lib.showTsos()
    print("Process Complete")
    return


# CLOSING


def saveLipd(filename):
    """
    Saves changes made to the target LiPD file.
    (ex. saveLiPD NAm-ak000.lpd)
    :param filename: (str) LiPD filename
    """
    lipd_lib.saveLipd(filename)
    print("Process Complete")
    return


def saveLipds():
    """
    Save changes made to all LiPD files in the workspace.
    """
    lipd_lib.saveLipds()
    print("Process Complete")
    return


def removeLipd(filename):
    """
    Remove LiPD object from library
    :return: None
    """
    lipd_lib.removeLipd(filename)
    print("Process Complete")
    return


def removeLipds():
    """
    Remove all LiPD objects from library.
    :return: None
    """
    lipd_lib.removeLipds()
    print("Process Complete")
    return


def quit():
    """
    Quit and exit the program. (Does not save changes)
    """
    # self.llib.close()
    print("Quitting...")
    return True


def set_source():
    """
    User sets the path to LiPD source. Local or online.
    :return: (str) Path
    """
    _path = ''
    invalid = True
    count = 0
    while invalid:
        print("Where are your files stored?\n1. Online URL\n2. Browse Computer\n3. "
              "Downloads folder\n4. Notebooks folder\n")
        option = input("Option: ")
        if option == '1':
            # Retrieve data from the online URL
            _path = input("Enter the URL: ")
        elif option == '2':
            # Open up the GUI browse dialog
            _path = browse_dialog()
            # Set the path to the local files in CLI and lipd_lib
        elif option == '3':
            # Set the path to the system downloads folder.
            _path = os.path.expanduser('~/Downloads')
        elif option == '4':
            # Set path to the Notebook folder
            _path = os.path.expanduser('~/LiPD_Notebooks')
        else:
            # Something went wrong. Prompt again. Give a couple tries before defaulting to downloads folder
            if count == 2:
                print("Defaulting to Downloads Folder.")
                _path = os.path.expanduser('~/Downloads')
            else:
                count += 1
                print("Invalid option. Try again.")
        if _path:
            invalid = False
    lipd_lib.setDir(_path)

    return _path


# GLOBALS
lipd_lib = LiPD_Library()
ts_lib = TimeSeries_Library()
convert = Convert()
path = ''
set_source()




from copy import deepcopy
import re
import csv

from ..helpers.google import get_google_csv
from ..helpers.directory import check_file_age

# LiPD to TIME SERIES

# GLOBALS
EMPTY = ['', ' ', None, 'na', 'n/a', 'nan', '?']

re_misc_fetch = re.compile(r'(geo_(\w+)|climateInterpretation_(\w+)|calibration_(\w+)|paleoData_(\w+))')
re_pub_fetch = re.compile(r'pub1_(citation|year|DOI|author|publisher|title|type|volume|issue|journal|link|pubDataUrl|abstract|pages)')

re_pub_valid = re.compile(r'pub(\d)_(citation|year|DOI|author|publisher|title|type|volume|issue|journal|link|pubDataUrl|abstract|pages)')
re_fund_valid = re.compile(r'funding(\d)_(grant|agency)')

re_pub_invalid = re.compile(r'pub_(\w+)|pub(\d)_(\w+)|pub(\d)(\w+)|pub(\w+)')
re_fund_invalid = re.compile(r'agency|grant|funding_agency|funding_grant')
re_geo_invalid = re.compile(r'geo(\w+)|geo_(\w+)')
re_paleo_invalid = re.compile(r'paleodata(\w+)|paleodata_(\w+)|measurement(\w+)|measurement_(\w+)')
re_calib_invalid = re.compile(r'calibration(\w+)|calibration_(\w+)')
re_clim_invalid = re.compile(r'climateinterpretation(\w+)|climateinterpretation_(\w+)')

re_pub_nh = re.compile(r'pub(\d)_(\w+)')
re_pub_cc = re.compile(r'pub(\w+)')
re_pub_h = re.compile(r'pub_(\w+)')
re_pub_n = re.compile(r'pub(\d)(\w+)')


class Convert(object):

    def __init__(self):

        # LiPD to TS (One file at a time)
        self.ts_root = {}  # Root metadata for current LiPD.
        self.ts_tsos = {}  # Individual columns. One entry represents one TSO (to be created later)

        # TS to LiPD (Batch Process)
        self.lipd_tsos = []  # One entry for each TSO metadata dictionary
        self.lipd_master = {}  # Key: LiPD name, Value: Complete metadata dictionary (when finished)
        self.lipd_curr_tso_data = {}  # Holds metadata for current TSO being processed
        self.dataset_ext = ''  # Filename.lpd. This is what the output name will be.
        self.table = ''
        self.variableName = ''

        # TS Names validation
        self.full_list = {"root": [], "pub": [], "climateInterpretation": [], "calibration": [], "geo": [],
                          "paleoData": []}
        # full_list - All valid TS Names and synonyms. { 'category' : [ ['validTSName', 'synonym'], ... ] }
        self.quick_list = []  # All valid TS Name keys
        self.recent = {}  # Recently corrected TS names, builds with each run.

    # LiPD TO TIME SERIES

    def ts_extract_main(self, d):
        """
        Main function to initiate LiPD to TSOs conversion.
        :param d: (dict) Metadata for one LiPD file
        """
        # Reset these each run
        self.ts_root = {}
        self.ts_tsos = {}

        # Build the root level data.
        # This will serve as the template for which column data will be added onto later.
        for k, v in d.items():
            if isinstance(v, str):
                self.ts_root[k] = v
            elif k == "funding":
                self.__ts_extract_funding(v)
            elif k == "geo":
                self.__ts_extract_geo(v)
            elif k == 'pub':
                self.__ts_extract_pub(v)

        # Create tso dictionaries for each individual column (build on root data)
        self.__ts_extract_paleo(d)

        # Get TSNames and verify against TSOs in ts_tsos
        self.__fetch_tsnames()
        self.__verify_tsnames()

        return self.ts_tsos

    def __ts_extract_funding(self, l):
        """
        Creates flat funding dictionary.
        :param l: (list) Funding entries
        """
        for idx, i in enumerate(l):
            for k, v in i.items():
                self.ts_root['funding' + str(idx+1) + '_' + k] = v
        return

    def __ts_extract_geo(self, d):
        """
        Extract geo data from input
        :param d: (d) Geo dictionary
        """
        # May not need these if the key names are corrected in the future.
        x = ['geo_meanLat', 'geo_meanLon', 'geo_meanElev']
        # Iterate through geo dictionary
        for k, v in d.items():
            # Case 1: Coordinates special naming
            if k == 'coordinates':
                for idx, p in enumerate(v):
                    try:
                        # Check that our value is not in EMPTY.
                        if isinstance(p, str):
                            if p.lower() in EMPTY:
                                # If elevation is a string or 0, don't record it
                                if idx != 2:
                                    # If long ot lat is empty, set it as 0 instead
                                    self.ts_root[x[idx]] = 0
                            else:
                                # Set the value as a float into its entry.
                                self.ts_root[x[idx]] = float(p)
                        # Value is a normal number, or string representation of a number
                        else:
                            # Set the value as a float into its entry.
                            self.ts_root[x[idx]] = float(p)

                    except IndexError:
                        continue
            # Case 2: Any value that is a string can be added as-is
            elif isinstance(v, str):
                if k == 'meanElev':
                    try:
                        # Some data sets have meanElev listed under properties for some reason.
                        self.ts_root['geo_' + k] = float(v)
                    except ValueError:
                        # If the value is a string, then we don't want it
                        continue
                else:
                    self.ts_root['geo_' + k] = v
            # Case 3: Nested dictionary. Recursion
            elif isinstance(v, dict):
                self.__ts_extract_geo(v)

        return

    def __ts_extract_pub(self, l):
        """
        Extract publication data from one or more publication entries.
        :param l: (list) Publication list
        """
        # For each publication entry
        for idx, pub in enumerate(l):
            # Get author data first, since that's the most ambiguously structured data.
            self.__ts_extract_authors(pub, idx)

            # Go through data of this publication
            for k, v in pub.items():
                # Case 1: DOI ID. Don't need the rest of 'identifier' dict
                if k == 'identifier':
                    try:
                        self.ts_root['pub' + str(idx+1) + '_DOI'] = v[0]['id']
                    except KeyError:
                        continue
                # Case 2: All other string entries
                else:
                    if k != 'authors' and k != 'author':
                        self.ts_root['pub' + str(idx+1) + '_' + k] = v

        return

    def __ts_extract_authors(self, pub, idx):
        """
        Create a concatenated string of author names. Separate names with semi-colons.
        :param pub: (unknown type) Publication author structure is ambiguous
        :param idx: (int) Index number of Pub
        """
        try:
            # DOI Author data. We'd prefer to have this first.
            names = pub['author']
        except KeyError:
            try:
                # Manually entered author data. This is second best.
                names = pub['authors']
            except KeyError:
                # Couldn't find any author data. Skip it altogether.
                names = False

        # If there is author data, find out what type it is
        if names:
            # Build author names onto empty string
            auth = ''
            # Is it a list of dicts or a list of strings? Could be either
            # Authors: Stored as a list of dictionaries or list of strings
            if isinstance(names, list):
                for name in names:
                    if isinstance(name, str):
                        auth += name + ';'
                    elif isinstance(name, dict):
                        for k, v in name.items():
                            auth += v + ';'
            elif isinstance(names, str):
                auth = names
            # Enter finished author string into target
            self.ts_root['pub' + str(idx+1) + '_author'] = auth[:-1]
        return

    def __ts_extract_paleo(self, d):
        """
        Extract all data from a PaleoData dictionary.
        :param d: (dict) PaleoData dictionary
        """
        # For each table in paleoData
        for k, v in d['paleoData'].items():
            # Get root items for this table
            self.__ts_extract_paleo_table_root(v)
            # Start creating TSOs with dictionary copies.
            for i, e in v['columns'].items():
                # TSO. Add this column onto root items. Deepcopy since we need to reuse ts_root
                col = self.__ts_extract_paleo_columns(e, deepcopy(self.ts_root))
                self.ts_tsos[d['dataSetName'] + '_' + k + '_' + i] = col
        return

    def __ts_extract_paleo_table_root(self, d):
        """
        Extract data from the root level of a paleoData table.
        :param d: (dict) One paleoData table
        """
        for k, v in d.items():
            if isinstance(v, str):
                self.ts_root['paleoData_' + k] = v
        return

    def __ts_extract_paleo_columns(self, d, tmp_tso):
        """
        Extract data from one paleoData column
        :param d: (dict) Column dictionary
        :param tmp_tso: (dict) TSO dictionary with only root items
        :return: (dict) Finished TSO
        """
        for k, v in d.items():
            if k == 'climateInterpretation':
                tmp_tso = self.__ts_extract_climate(v, tmp_tso)
            elif k == 'calibration':
                tmp_tso = self.__ts_extract_calibration(v, tmp_tso)
            else:
                # Assume if it's not a special nested case, then it's a string value
                tmp_tso['paleoData_' + k] = v
        return tmp_tso

    @staticmethod
    def __ts_extract_calibration(d, tmp_tso):
        """
        Get calibration info from column data.
        :param d: (dict) Calibration dictionary
        :param tmp_tso: (dict) Temp TSO dictionary
        :return: (dict) tmp_tso with added calibration entries
        """
        for k, v in d.items():
            tmp_tso['calibration_' + k] = v
        return tmp_tso

    @staticmethod
    def __ts_extract_climate(d, tmp_tso):
        """
        Get climate interpretation from column data.
        :param d: (dict) Climate Interpretation dictionary
        :param tmp_tso: (dict) Temp TSO dictionary
        :return: (dict) tmp_tso with added climateInterpretation entries
        """
        for k, v in d.items():
            tmp_tso['climateInterpretation_' + k] = v
        return tmp_tso

    # TIME SERIES to LiPD

    def lipd_extract_main(self, lipd_tsos):
        """
        Main function to initiate TimeSeries to LiPD conversion
        :param lipd_tsos: (list of dict) List of all TSOs
        :return:
        """
        # Reset for each run
        self.lipd_tsos = lipd_tsos  # All TSO metadata dictionaries
        self.lipd_master = {}  # All lipds (in progress) by dataset name. Key: LiPD name, Val: JSON metadata
        self.lipd_curr_tso_data = {}  # Current TSO metadata

        # Receive list of TSO objects
        for tso in self.lipd_tsos:
            # Set current keys
            self.dataset_ext = tso['name']
            self.lipd_curr_tso_data = tso['data']
            self.__lipd_set_current()

            # Since root items are the same in the same dataset, we only need these steps if dataset doesn't exist yet.
            if self.dataset_ext not in self.lipd_master:
                self.__lipd_extract_roots()

            # Extract PaleoData, Calibration, and Climate Interpretation
            self.__lipd_extract_paleo()

        return self.lipd_master

    def __lipd_set_current(self):
        """
        Set keys for current TSO file
        """
        # Get key info
        try:
            self.table = self.lipd_curr_tso_data['paleoData_paleoDataTableName']
            self.variableName = self.lipd_curr_tso_data['paleoData_variableName']
        except KeyError:
            print("KeyError: Initial keys")
        return

    def __lipd_extract_roots(self):
        """
        Extract root from TimeSeries
        """
        tmp_funding = {}
        tmp_pub = {}
        tmp_master = {'pub': [], 'geo': {'geometry': {'coordinates': []}, 'properties': {}}, 'funding': [],
                      'paleoData': {}}
        p_keys = ['siteName', 'pages2kRegion']
        c_keys = ['meanLat', 'meanLon', 'meanElev']
        c_vals = [0, 0, 0]

        for k, v in self.lipd_curr_tso_data.items():
            if 'paleoData' not in k and 'calibration' not in k and 'climateInterpretation' not in k:
                if 'funding' in k:
                    # Group funding items in tmp_funding by number
                    m = re_fund_valid.match(k)
                    try:
                        tmp_funding[m.group(1)][m.group(2)] = v
                    except KeyError:
                        tmp_funding[m.group(1)] = {}
                        tmp_funding[m.group(1)][m.group(2)] = v

                elif 'geo' in k:
                    key = k.split('_')
                    # Coordinates
                    if key[1] in c_keys:
                        if key[1] == 'meanLat':
                            c_vals[0] = v
                        elif key[1] == 'meanLon':
                            c_vals[1] = v
                        elif key[1] == 'meanElev':
                            c_vals[2] = v
                    # Properties
                    elif key[1] in p_keys:
                        tmp_master['geo']['properties'][key[1]] = v
                    # All others
                    else:
                        tmp_master['geo'][key[1]] = v

                elif 'pub' in k:
                    # Group pub items in tmp_pub by number
                    m = re_pub_valid.match(k)
                    number = int(m.group(1)) - 1  # 0 indexed behind the scenes, 1 indexed to user.
                    key = m.group(2)
                    # Authors ("Pu, Y.; Nace, T.; etc..")
                    if key == 'author' or key == 'authors':
                        try:
                            tmp_pub[number]['author'] = self.__lipd_extract_author(v)
                        except KeyError:
                            tmp_pub[number] = {}
                            tmp_pub[number]['author'] = self.__lipd_extract_author(v)
                    # DOI ID
                    elif key == 'DOI':
                        try:
                            tmp_pub[number]['identifier'] = [{"id": v, "type": "doi", "url": "http://dx.doi.org/" + str(v)}]
                        except KeyError:
                            tmp_pub[number] = {}
                            tmp_pub[number]['identifier'] = [{"id": v, "type": "doi", "url": "http://dx.doi.org/" + str(v)}]
                    # All others
                    else:
                        try:
                            tmp_pub[number][key] = v
                        except KeyError:
                            tmp_pub[number] = {}
                            tmp_pub[number][key] = v
                else:
                    # Root
                    tmp_master[k] = v

        # Append pub and funding to master
        for k, v in tmp_pub.items():
            tmp_master['pub'].append(v)
        for k, v in tmp_funding.items():
            tmp_master['funding'].append(v)
        # Get rid of elevation coordinate if one was never added.
        if c_vals[2] == 0:
            del c_vals[2]
        tmp_master['geo']['geometry']['coordinates'] = c_vals

        # If we're getting root info, then there shouldn't be a dataset entry yet.
        # Create entry in object master, and set our new data to it.
        self.lipd_master[self.dataset_ext] = {}
        self.lipd_master[self.dataset_ext] = tmp_master

        return

    @staticmethod
    def __lipd_extract_author(s):
        """
        Split author string back into organized dictionary
        :param s: (str) Formatted names string "Last, F.; Last, F.; etc.."
        :return: (list of dictionaries) One dictionary per author name
        """
        l = []
        authors = s.split(';')
        for author in authors:
            l.append({'name': author})
        return l

    def __lipd_extract_paleo_root(self):
        """
        Create paleo table in master if it doesn't exist. Insert table root items
        """

        try:
            # If table doesn't exist yet, create it.
            if self.table not in self.lipd_master[self.dataset_ext]['paleoData']:
                t_root_keys = ['filename', 'googleWorkSheetKey', 'paleoDataTableName']
                self.lipd_master[self.dataset_ext]['paleoData'][self.table] = {'columns': {}}
                for key in t_root_keys:
                    try:
                        # Now set the root table items in our new table
                        self.lipd_master[self.dataset_ext]['paleoData'][self.table][key] = self.lipd_curr_tso_data['paleoData_' + key]
                    except KeyError:
                        # That's okay. Keep trying :)
                        continue
        except KeyError:
            print("KeyError lipd_extract_paleo: Inserting table ")

        return

    def __lipd_extract_paleo(self):
        """
        Extract paleoData from tso
        :return: LiPD entry in self.lipd_master is updated
        """
        tmp_clim = {}
        tmp_cal = {}

        self.__lipd_extract_paleo_root()

        try:
            # Create the column entry in the table
            self.lipd_master[self.dataset_ext]['paleoData'][self.table]['columns'][self.variableName] = {}
            # Start inserting paleoData into the table at self.lipd_master[table][column]
            for k, v in self.lipd_curr_tso_data.items():
                # ['paleoData', 'key']
                m = k.split('_')
                if 'paleoData' in m[0]:
                    self.lipd_master[self.dataset_ext]['paleoData'][self.table]['columns'][self.variableName][m[1]] = v
                elif 'calibration' in m[0]:
                    tmp_cal[m[1]] = v
                elif 'climateInterpretation' in m[0]:
                    tmp_clim[m[1]] = v

        except KeyError:
            print("KeyError lipd_extract_paleo: Inserting column data ")

        # If these sections had any items added to them, then add them to the column master.
        if tmp_clim:
            self.lipd_master[self.dataset_ext]['paleoData'][self.table]['columns'][self.variableName]['climateInterpretation'] = tmp_clim
        if tmp_cal:
            self.lipd_master[self.dataset_ext]['paleoData'][self.table]['columns'][self.variableName]['calibration'] = tmp_cal
        return

    # VALIDATING AND UPDATING TSNAMES

    def __fetch_tsnames(self):
        """
        Call down a current version of the TSNames spreadsheet from google. Convert to a structure better
        for comparisons.
        :return: (self.full_list) Keys: Valid TSName, Values: TSName synonyms
        :return: (self.quick_list) List of valid TSnames
        """

        # Check if it's been longer than one day since updating the TSNames.csv file.
        # If so, go fetch the file from google in case it's been updated since.
        # Or if file isn't found at all, download it also.
        if check_file_age('tsnames.csv', 1):
            # Fetch TSNames sheet from google
            print("Fetching update for TSNames.csv")
            ts_id = '1C135kP-SRRGO331v9d8fqJfa3ydmkG2QQ5tiXEHj5us'
            get_google_csv(ts_id, 'tsnames.csv')
        try:
            # Start sorting the tsnames into an organized structure
            with open('tsnames.csv', 'r') as f:
                r = csv.reader(f, delimiter=',')
                for idx, line in enumerate(r):
                    # print('line[{}] = {}'.format(i, line))
                    if idx != 0:
                        # Do not record empty lines. Create list of non-empty entries.
                        line = [x for x in line if x]
                        # If line has content (i.e. not an empty line), then record it
                        if line:
                            # We don't need all the duplicates of pub and fund.
                            if "pub" in line[0] or "funding" in line[0]:
                                if re_pub_fetch.match(line[0]):
                                    self.quick_list.append(line[0])
                                    self.full_list['pub'].append(line)
                            elif re_misc_fetch.match(line[0]):
                                # Other Categories. Not special
                                self.quick_list.append(line[0])
                                cat, key = line[0].split('_')
                                self.full_list[cat].append(line)
                            else:
                                # Any of the root items
                                self.quick_list.append(line[0])
                                self.full_list["root"].append(line)
        except FileNotFoundError:
            print("CSV FileNotFound: TSNames")

        return

    def __verify_tsnames(self):
        """
        Verify TSNames are current and valid. Compare to TSNames spreadsheet in Google Drive. Update where necessary.
        """
        # Temp to store incorrect keys
        bad_keys = []

        for name, tso in self.ts_tsos.items():
            # Build onto the "recent" dictionary so we have a list of keys to replace.
            for k, v in tso.items():
                # @context needs to be ignored
                if k not in self.quick_list and not re_pub_valid.match(k) and not re_fund_valid.match(k) and k != '@context':
                    # Invalid key. Store in temp for processing.
                    if k not in self.recent:
                        bad_keys.append(k)
            # Start to find replacements for empty entries in "recent"
            for incorrect in bad_keys:
                # Set incorrect name as key, and valid name as value.
                self.recent[incorrect] = self.__get_valid_tsname(incorrect)

            # Use temp to start replacing entries in d
            for invalid, valid in self.recent.items():
                try:
                    # Add new item, and remove old item in one step
                    tso[valid] = tso.pop(invalid)
                except KeyError:
                    continue

        return

    def __get_valid_tsname(self, invalid):
        """
        Turn a bad tsname into a valid one.
        * Note: Index[0] for each TSName is the most current, valid entry. Index[1:] are synonyms
        :param invalid: (str) Invalid tsname
        :return: (str) valid tsname
        """
        valid = ''
        invalid = invalid.lower()
        try:
            # PUB ENTRIES
            if re_pub_invalid.match(invalid):

                # Case 1: pub1_year (number and hyphen)
                if re_pub_nh.match(invalid):
                    s_invalid = invalid.split('_', 1)
                    # Check which list the key word is in
                    for line in self.full_list['pub']:
                        for key in line:
                            if s_invalid[1] in key.lower():
                                # Get the keyword from the valid entry.
                                v = line[0].split("_")
                                # Join our category with the valid keyword
                                valid = ''.join([s_invalid[0], '_', v[1]])

                # Case 2: pub_year (hyphen)
                elif re_pub_h.match(invalid):
                    s_invalid = invalid.split('_', 1)
                    # The missing pub number is the main problem, but the keyword may or may not be correct. Check.
                    for line in self.full_list['pub']:
                        for key in line:
                            if s_invalid[1] in key.lower():
                                # We're going to use the valid entry as-is, because that's what we need for this case.
                                valid = line[0]

                # Case 3: pub1year (number)
                elif re_pub_n.match(invalid):
                    s_invalid = re_pub_n.match(invalid)
                    for line in self.full_list['pub']:
                        for key in line:
                            if s_invalid.group(2) in key.lower():
                                v = line[0].split('_', 1)
                                valid = ''.join(['pub', s_invalid.group(0), v[1]])

                # Case 4: pubYear (camelcase)
                elif re_pub_cc.match(invalid):
                    valid = self.__iter_ts('pub', invalid)

            # FUNDING
            elif re_fund_invalid.match(invalid):
                if "grant" in invalid:
                    valid = 'funding1_grant'
                elif "agency" in invalid:
                    valid = "funding1_agency"

            # GEO
            elif re_geo_invalid.match(invalid):
                valid = self.__iter_ts('geo', invalid)

            # PALEODATA
            elif re_paleo_invalid.match(invalid):
                g1 = re_paleo_invalid.match(invalid).group(1)
                valid = self.__iter_ts('paleoData', g1)

            # CALIBRATION
            elif re_calib_invalid.match(invalid):
                g1 = re_calib_invalid.match(invalid).group(1)
                valid = self.__iter_ts('calibration', g1)

            # CLIMATE INTERPRETATION
            elif re_clim_invalid.match(invalid):
                g1 = re_clim_invalid.match(invalid).group(1)
                if 'climate' in g1:
                    g1 = re.sub('climate', '', g1)
                valid = self.__iter_ts('climateInterpretation', g1)

            else:
                # ROOT
                valid = self.__iter_ts('root', invalid)

            # LAST CHANCE:
            # Specific case that isn't a typical format, or no match. Go through all possible entries.
            if not valid:
                valid = self.__iter_ts(None, invalid)

        except IndexError:
            print("Get TSName: Something went wrong")

        if not valid:
            print("TSName: Couldn't find a match: " + invalid)

        return valid

    def __iter_ts(self, category, invalid):
        """
        Match an invalid entry to one of the TSName synonyms.
        :param category: (str) Name of category being searched
        :param invalid: (str) Invalid tsname string
        :return: (str) Valid tsname
        """
        valid = ''

        # If a leading hyphen is in the string, get rid of it.
        if '_' == invalid[0]:
            invalid = invalid[1:]

        # If one specific category is passed through
        if category:
            for line in self.full_list[category]:
                for key in line:
                    if invalid in key.lower():
                        valid = line[0]
                        break
        # If the entire TSNames dict is passed through (i.e. final effort, all categories have failed so far)
        else:
            for k, v in self.full_list.items():
                for line in v:
                    for key in line:
                        if invalid in key.lower():
                            valid = line[0]
                            break

        return valid

    # HELPERS

    # def create_tso(self):
    #     """
    #     Creates a TimeSeriesObject and add it to the TimeSeriesLibrary
    #     :return: (obj) Time Series Object
    #     """
    #     for name, tso in self.ts_tsos.items():
    #         TimeSeries().load(tso)
    #     return



from .TimeSeries import TimeSeries


class TimeSeries_Library(object):
    def __init__(self):
        self.master = {}  # Master data. (Key: Name of TimeSeries Object. Value: TimeSeries Object)

    # LOADING

    def loadTso(self, name, d):
        """
        Load in a single TimeSeries Objects.
        :param name: (str) TSO name
        :param d: (dict) TSO metadata
        """
        t = TimeSeries()
        t.load(d)
        self.master[name] = t
        return

    def loadTsos(self, name_ext, d):
        """
        Load in multiple TimeSeries Objects.
        :param name_ext: (str) Need lpd filename in case we have to convert TSOs back to LiPDs.
        :param d: (dict) All TSOs resulting from one LiPD file. K: TS names, V: TS metadata
        """
        # Create a TSO for each, and load into the TS_Library master
        for k, v in d.items():
            t = TimeSeries()
            t.load(v)
            t.set_ts_name(k)
            t.set_lpd_name(name_ext)
            self.master[k] = t
        return

    # ANALYSIS

    def showTso(self, name):
        """
        Show contents of one TSO object.
        :param name:
        :return:
        """
        try:
            self.master[name].display_data()
        except KeyError:
            print("TimeSeries not found")
        return

    def showTsos(self):
        """
        Display all TimeSeries names in the TimeSeries_Library
        :return:
        """
        for name, tso in sorted(self.master.items()):
            print(name)
        return

    def saveTso(self, tso):
        """
        Write data for one specified TSO to file.
        :param tso: (obj)
        :return:
        """
        pass

    def saveTsos(self):
        """
        Write data from all TSOs to file.
        :return:
        """
        pass

    # CLOSING

    # HELPERS

    def get_master(self):
        """
        Get master list of TSO names and TSOs
        :return: (dict) Master dictionary
        """
        return self.master




import string

from .LiPD import LiPD
from ..helpers.directory import *
from ..helpers.google import *


class LiPD_Library(object):
    """
    The LiPD Library is meant to encompass a collection of LiPD file objects that are being analyzed in the current
    workspace. The library holds one LiPD object for each LiPD file that is loaded.
    """

    def __init__(self):
        self.dir_root = ''
        self.dir_tmp = create_tmp_dir()
        self.master = {}

    # LOADING

    def setDir(self, dir_root):
        """
        Changes the current working directory.
        :param dir_root:
        :return:
        """
        try:
            self.dir_root = dir_root
            os.chdir(self.dir_root)
            # print("Directory: " + str(dir_root) + '\n')
        except FileNotFoundError:
            pass
            # print("Not a valid directory")
        return

    def loadLipd(self, name):
        """
        Load a single LiPD object into the LiPD Library.
        :param name: (str) Filename
        :return: None
        """
        self.__append_lipd(name)
        print("Loaded 1 LiPD file")
        return

    def loadLipds(self):
        """
        Load a directory (multiple) LiPD objects into the LiPD Library
        :return:
        """
        # Confirm that a CWD is set first.
        if self.dir_root == '':
            print("Current Working Directory has not been set.")
            return
        os.chdir(self.dir_root)
        # Get a list of all lpd files
        file_list = list_files('lpd')
        # Loop: Append each file to Library
        try:
            for name_ext in file_list:
                self.__append_lipd(name_ext)
            print("Loaded {0} LiPD file(s)".format(len(file_list)))
        except:
            print("Error loading files")

        return

    # ANALYSIS

    def showCsv(self, name):
        """
        Show CSV data from one LiPD object
        :param name: (str) Filename
        :return:
        """
        try:
            self.master[name].display_csv()
        except KeyError:
            print("LiPD not found")
        return

    def showMetadata(self, name):
        """
        Display data from target LiPD file.
        :param name: (str) Filename
        :return:
        """
        try:
            self.master[name].display_json()
        except KeyError:
            print("LiPD not found")
        return

    def showLipdMaster(self, name):
        """
        Display data from target LiPD file.
        :param name: (str) Filename
        :return:
        """
        try:
            self.master[name].display_master()
        except KeyError:
            print("LiPD not found")
        return

    def showLipds(self):
        """
        Display all LiPD files in the LiPD Library
        :return:
        """
        print("Found: {} file(s)".format(len(self.master)))
        for k, v in sorted(self.master.items()):
            print(k)
        return

    def showMap(self, filename):
        """
        Intermediate function. Determines which type of map to create based on input.
        :param filename: (str) Map the specified LiPD files. (none) Map all files in LiPD_Library.
        :return: (img) Google map with location markers
        """
        # No input given. Map all LiPDs
        if not filename:
            self.map_all()
        # One or more records given. Map them.
        else:
            self.map_some(filename)
        return

    def map_some(self, files):
        """
        Map one or more specified LiPDs.
        :param files: (str) Comma separated filenames
        :return: None
        """
        f = files.split(',')
        chars = list(string.ascii_uppercase + string.digits)
        markers = []
        # Call a single point
        if len(f) is 1:
            c = self.master[f[0]].data_json['geo']['geometry']['coordinates']
            c_str = str(c[0]) + ',' + str(c[1])
            markers.append("markers=size:large|color:red|label:1|" + c_str)
            get_static_google_map(f[0], center=c_str, zoom=7, markers=markers)

        # Call a list of markers
        else:
            # Pull coordinates from master list for each given filename
            for idx, filename in enumerate(f):
                try:
                    c = self.master[filename].data_json['geo']['geometry']['coordinates']
                    markers.append('markers=size:large|color:red|label:' + chars[idx] + '|' + str(c[0]) + ',' + str(c[1]))
                    print('Key: ' + chars[idx] + '  Record: ' + str(filename))

                except KeyError:
                    print('Coordinates Error: ' + str(filename))

            # Map the list of markers
            get_static_google_map('multi-marker', markers=markers)
        return

    def map_all(self):
        """
        Map all LiPDs in the library.
        """
        chars = list(string.ascii_uppercase + string.digits)
        markers = []
        n = 0

        # Pull coordinates from master list for each given filename
        for k, v in self.master.items():
            try:
                c = v.data_json['geo']['geometry']['coordinates']
                markers.append('markers=size:large|color:red|label:' + chars[n] + '|' + str(c[0]) + ',' + str(c[1]))
                print('Key: ' + chars[n] + '  Record: ' + str(k))
                n += 1
            except KeyError:
                print("Coordinates Error: " + str(k))
        # Map the list of markers
        get_static_google_map('multi-marker', markers=markers)
        return

    # CLOSING

    def saveLipd(self, name):
        """
        Overwrite LiPD files in OS with LiPD data in the current workspace.
        """
        try:
            self.master[name].save()
        except KeyError:
            print("LiPD not found")
        return

    def saveLipds(self):
        """
        Overwrite target LiPD file in OS with LiPD data in the current workspace.
        """
        for k, v in self.master.items():
            self.master[k].save()

    def removeLipd(self, name):
        """
        Removes target LiPD file from the workspace. Delete tmp folder, then delete object.
        :param name: (str) Filename
        """
        try:
            self.master[name].remove()
            del self.master[name]
        except KeyError:
            print("LiPD not found")
        return

    def removeLipds(self):
        """
        Clear the workspace. Empty the master dictionary.
        """
        self.master = {}
        return

    # HELPERS

    def __append_lipd(self, name_ext):
        """
        Creates and adds a new LiPD object to the LiPD Library for the given LiPD file...
        :param name_ext: (str) Filename with extension
        """
        os.chdir(self.dir_root)
        # create a lpd object
        lipd_obj = LiPD(self.dir_root, self.dir_tmp, name_ext)
        # load in the data from the lipd file (unpack, and create a temp workspace)
        lipd_obj.load()
        # add the lpd object to the master dictionary
        self.master[name_ext] = lipd_obj
        return

    def get_master(self):
        """
        Retrieve the LiPD_Library master list. All names and LiPD objects.
        :return:
        """
        return self.master

    def load_tsos(self, d):
        """
        Overwrite converted TS metadata back into its matching LiPD object.
        :param d: (dict) Metadata from TSO
        """

        for name_ext, metadata in d.items():
            # Important that the dataSetNames match for TSO and LiPD object. Make sure
            try:
                self.master[name_ext].load_tso(metadata)
            except KeyError:
                print("LiPD_Library: Error loading " + str(name_ext) + " from TSO")

        return



from ..helpers.jsons import *


class TimeSeries(object):
    def __init__(self):
        self.master = {}  # Master data. JSON metadata
        self.ts_name = ''  # TimeSeries Object name. <Dataset_table_column> format.
        self.lpd_name = ''  # LiPD filename. Original from LiPD object.

    # LOADING

    def load(self, d):
        """
        Load TimeSeries metadata into object
        :param d: (dict) TS Metadata
        """
        self.master = d
        return

    def set_ts_name(self, ts_name):
        """
        Set the filename to match LiPD filename counterpart
        :param ts_name: (str) TimeSeries name. Dataset + table + column
        """
        self.ts_name = ts_name
        return

    def set_lpd_name(self, lpd_name):
        """
        Set the data set name for this TSO
        :param lpd_name: (str) Name_ext from the LiPD object. Needed if converting back to LiPD.
        """
        self.lpd_name = lpd_name
        return

    # ANALYSIS

    def get_master(self):
        """
        Retrieves the metadata from self.master
        :return: (dict) Metadata
        """
        return self.master

    def get_lpd_name(self):
        """
        Get LiPD filename for this dataset.
        """
        return self.lpd_name

    def display_data(self):
        """
        Displays the metadata contents of self.master
        """
        print(json.dumps(self.master, indent=2))
        return

    # CLOSING

    # HELPERS



import re
import os
import shutil

# Pre-compiled Regexes
first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')

# GLOBALS
# 13 is a list of keys to ignore when using create_blanks
SECTIONS = {1: ['onlineResource', 'studyName', 'archive', 'parameterKeywords', 'originalSourceUrl'],
            2: ['date'],
            3: ['studyName'],
            4: ['investigators'],
            5: ['description'],
            6: ['pub'],
            7: ['funding', 'agency', 'grant'],
            8: ['geo'],
            9: ['collectionName', 'earliestYear', 'mostRecentYear', 'timeUnit', 'coreLength', 'notes'],
            10: ['speciesName', 'commonName', 'treeSpeciesCode'],
            11: ['chronology'],
            12: ['paleoData'],
            13: ['funding', 'type', 'bbox', 'geo']}

# The order of the items in the list is the order that we want to write them to the file.
# 11 is the order for writing each column in the variables section
ORDERING = {1: ['studyName', 'onlineResource', 'originalSourceUrl', 'archive', 'parameterKeywords'],
            2: ['date'],
            3: ['studyName'],
            4: ['investigators'],
            5: ['description'],
            6: ['authors', 'publishedDateOrYear', 'publishedTitle', 'journalName', 'volume', 'edition', 'issue',
                'pages', 'doi', 'onlineResource', 'fullCitation', 'abstract', 'identifier'],
            7: ['agency', 'grant'],
            8: ['siteName', 'location', 'country', 'northernmostLatitude', 'southernmostLatitude',
                'easternmostLongitude', 'westernmostLongitude', 'elevation'],
            9: ['collectionName', 'earliestYear', 'mostRecentYear', 'timeUnit', 'coreLength', 'notes'],
            10: ['speciesName', 'commonName'],
            11: ['parameter', 'description', 'material', 'error', 'units', 'seasonality', 'archive', 'detail',
                 'method', 'dataType']}


class LPD_NOAA(object):
    """
    Creates a NOAA object that contains all the functions needed to write out a LiPD file as a NOAA text file.
    :return: (txt) NOAA File
    """

    def __init__(self, dir_root, name, root_dict):
        """
        :param dir_root: (str) Path to dir containing all .lpd files
        :param name: (str) Name of current .lpd file
        :param root_dict: (dict) Full dict loaded from jsonld file
        """
        self.dir_root = dir_root
        self.name = name
        self.root_dict = root_dict
        self.steps_dict = {1:{},2:{},3:{},4:{},5:{},6:{},7:{},8:{},9:{},10:{},11:{},12:{},13:{}}
        self.name_txt = name + '.txt'

    @staticmethod
    def __csv_found(filename):
        """
        Check for CSV file. Make sure it exists before we try to open it.
        :param filename: (str) Name of the file to open.
        :return: none
        """
        found = False

        # Current Directory
        # tmp/filename/data (access to jsonld, changelog, csv, etc)

        try:
            # Attempt to open csv
            if open(filename):
                found = True
                # print("{0} - found {1} csv".format(filename, datatype))
        except FileNotFoundError:
            # print("{0} - no {1} csv".format(filename, datatype))
            pass
        return found

    @staticmethod
    def __get_mv(d):
        """
        Get missing value from root of data table dictionary.
        :param d: (dict) Data table
        :return: (str) Missing value or empty str.
        """
        if 'missingValue' in d:
            return d['missingValue']
        return ''

    @staticmethod
    def __underscore(key):
        """
        Convert camelCase to underscore
        :param key: (str) Key or title name
        """

        # Special keys that need a specific key change
        if key == 'doi':
            key_o = 'DOI'

        elif key == 'agency':
            key_o = 'Funding_Agency_Name'

        elif key == 'originalSourceURL':
            key_o = 'Original_Source_URL'

        # Use regex to split and add underscore at each capital letter
        else:
            s = first_cap_re.sub(r'\1_\2', key)
            key_o = all_cap_re.sub(r'\1_\2', s).title()

        return key_o

    @staticmethod
    def __split_path(string):
        """
        Used in the path_context function. Split the full path into a list of steps
        :param string: (str) Path string ("geo-elevation-height")
        :return out: (list) Path as a list of strings. One entry per path step.(["geo", "elevation", "height"])
        """
        out = []
        position = string.find(':')
        if position != -1:
            # A position of 0+ means that ":" was found in the string
            key = string[:position]
            val = string[position+1:]
            out.append(key)
            out.append(val)
            if ('-' in key) and ('Funding' not in key) and ('Grant' not in key):
                out = key.split('-')
                out.append(val)
        return out

    @staticmethod
    def __get_corelength(d):
        """
        Get the value and unit to write the Core Length line
        :param d:
        :return val, unit:(int) Value (str) Unit
        """
        # If d is a string, it'll throw a type error
        if not isinstance(d, str):
            try:
                val = d['coreLength']['value']
            except KeyError:
                val = ''
            try:
                unit = d['coreLength']['unit']
            except KeyError:
                unit = ''
            return val, unit
        return '', ''

    @staticmethod
    def __get_identifier(d):
        """
        Get DOI id and url from identifier dictionary.
        :param d: (list) Identifier list. (Should only have one dictionary in it)
        :return: (str) DOI id, (str) DOI url
        """
        if not isinstance(d, str):
            try:
                doi_id = d[0]['id']
            except KeyError:
                doi_id = ''
            try:
                doi_url = d[0]['url']
            except KeyError:
                doi_url = ''
            return doi_id, doi_url
        return '', ''

    @staticmethod
    def __create_blanks(section_num, d):
        """
        All keys need to be written to the output, with or without a value. Furthermore, only keys that have values
        exist at this point. We need to manually insert the other keys with a blank value. Loop through the global list
        to see what's missing in our dict.
        :param section_num: (int) Retrieve data from global dict for this number.
        :param d: (dict) Section of steps_dict
        :return: none
        """
        for key in ORDERING[section_num]:
            if key not in d and key not in SECTIONS[13]:
                # Key not in our dict. Create the blank entry.
                d[key] = ''
        return

    @staticmethod
    def __get_filename(table):
        try:
            filename = table['filename']
        except KeyError:
            filename = ''
        return filename

    @staticmethod
    def __coordinates(l, d):
        """
        Reorganize coordinates based on how many values are available.
        :param l: (list of float) Coordinate values
        :param d: (dict) Location with cooresponding values
        :return:
        """

        length = len(l)
        locations = ['northernmostLatitude', 'easternmostLongitude', 'southernmostLatitude', 'westernmostLongitude']

        if length == 0:
            for location in locations:
                d[location] = ' '
        elif length == 2:
            d[locations[0]] = l[0]
            d[locations[1]] = l[1]
            d[locations[2]] = l[0]
            d[locations[3]] = l[1]

        elif length == 4:
            for index, location in enumerate(locations):
                d[locations[index]] = l[index]

        return d

    def main(self):
        """
        Load in the template file, and run through the parser
        :return:
        """

        # Starting Directory: dir_tmp/dir_bag/data/

        # NOAA files are organized in sections, but jld is not. Reorganize to match NOAA template.
        for k, v in self.root_dict.items():
            self.__reorganize(k, v)

        # Use data in steps_dict to write to
        self.__write_file()

        return

    def __write_file(self):
        """
        Open text file. Write one section at a time. Close text file. Move completed file to dir_root/noaa/
        :return: none
        """

        noaa_txt = open(self.name_txt, "w+")
        self.__write_top(noaa_txt, 1)
        self.__write_generic(noaa_txt, 'Contribution_Date', 2, self.steps_dict[2])
        self.__write_generic(noaa_txt, 'Title', 3, self.steps_dict[3])
        self.__write_generic(noaa_txt, 'Investigators', 4, self.steps_dict[4])
        self.__write_generic(noaa_txt, 'Description_Notes_and_Keywords', 5, self.steps_dict[5])
        self.__write_pub(noaa_txt, self.steps_dict[6]['pub'])
        self.__write_funding(noaa_txt, self.steps_dict[7])
        self.__write_geo(noaa_txt, self.steps_dict[8])
        self.__write_generic(noaa_txt, 'Data_Collection', 9, self.steps_dict[9])
        self.__write_generic(noaa_txt, 'Species', 10, self.steps_dict[10])
        # self.write_chron(noaa_txt, self.steps_dict[11])

        # Run once for each table. Keep variables and related data grouped
        for table in self.steps_dict[12]['paleoData']:
            self.__write_variables(noaa_txt, table)
            self.__write_numbers(noaa_txt, table)
        noaa_txt.close()
        shutil.copy(os.path.join(os.getcwd(), self.name_txt), self.dir_root)
        return

    def __write_top(self, noaa_txt, section_num):
        """
        Write the top section of the txt file.
        :param noaa_txt: (obj) Output .txt file that is being written to.
        :param section_num: (int) Section number
        :return: none
        """
        self.__create_blanks(section_num, self.steps_dict[section_num])
        noaa_txt.write(
            '# ' + self.steps_dict[section_num]['studyName'] +
            '\n#-----------------------------------------------------------------------\
            \n#                World Data Service for Paleoclimatology, Boulder\
            \n#                                  and\
            \n#                     NOAA Paleoclimatology Program\
            \n#             National Centers for Environmental Information (NCEI)\
            \n#-----------------------------------------------------------------------\
            \n# Template Version 2.0\
            \n# NOTE: Please cite Publication, and Online_Resource and date accessed when using these data.\
            \n# If there is no publication information, please cite Investigators, Title, and Online_Resource and date accessed.\
            \n# Online_Resource: ' + self.steps_dict[section_num]['onlineResource'] + '\
            \n#\
            \n# Original_Source_URL: ' + self.steps_dict[section_num]['originalSourceUrl'] + '\
            \n#\
            \n# Description/Documentation lines begin with #\
            \n# Data lines have no #\
            \n#\
            \n# Archive: ' + self.steps_dict[section_num]['archive'] + '\
            \n#\
            \n# Parameter_Keywords: ' + self.steps_dict[section_num]['parameterKeywords'] + '\
            \n#--------------------\n')

        return

    def __write_generic(self, noaa_txt, header, section_num, d):
        """
        Write a generic section to the .txt. This function is reused for multiple sections.
        :param noaa_txt: (obj) Output .txt file that is being written to.
        :param header: (str) Header title for this section
        :param section_num: (int)
        :param d: (dict) Section from steps_dict
        :return: none
        """
        self.__create_blanks(section_num, d)
        noaa_txt.write('# ' + header + ' \n')
        for entry in ORDERING[section_num]:
            if entry == 'coreLength':
                val, unit = self.__get_corelength(d[entry])
                noaa_txt.write('#   ' + self.__underscore(entry) + ': ' + str(val) + ' ' + str(unit) + '\n')
            elif entry == 'identifier':
                doi_id, doi_url = self.__get_identifier(d[entry])
                noaa_txt.write('#   DOI_id: ' + doi_id + '\n')
                noaa_txt.write('#   DOI_url: ' + doi_url + '\n')
            else:
                noaa_txt.write('#   ' + self.__underscore(entry) + ': ' + str(d[entry]) + '\n')
        noaa_txt.write('#------------------\n')
        return

    def __write_pub(self, noaa_txt, pub_root):
        """
        Write pub section. There may be multiple, so write a generic section for each one.
        :param noaa_txt: (obj) Output .txt file that is being written to.
        :param pub_root: (list) One or more pub dictionaries.
        :return: none
        """
        for pub in pub_root:
            self.__write_generic(noaa_txt, 'Publication', 6, pub)
        return

    def __write_funding(self, noaa_txt, d):
        """

        :param noaa_txt:
        :param d:
        :return:
        """
        for key, fund_list in d.items():
            for fund in fund_list:
                self.__write_generic(noaa_txt, 'Funding_Agency', 7, fund)
        return

    def __write_geo(self, noaa_txt, d):
        """

        :param noaa_txt:
        :param d:
        :return:
        """
        for k, v in d.items():
            d = self.__reorganize_geo(v)
        self.__write_generic(noaa_txt, 'Site_Information', 8, d)
        return

    def __write_chron(self, noaa_txt, d):
        """

        :param file_in:
        :param noaa_txt:
        :param d:
        :return:
        """
        noaa_txt.write('# Chronology:\n#\n')

        if self.__csv_found('chron'):
            cols = d['chronology']['columns']
            # Write variables line from dict_in
            for index, col in enumerate(cols):
                if index == 0:
                    noaa_txt.write('#       ' + col['parameter'] + ' (' + col['units'] + ')  | ')
                elif index == len(cols)-1:
                    noaa_txt.write(col['parameter'] + ' (' + col['units'] + ')\n')
                else:
                    noaa_txt.write(col['parameter'] + ' (' + col['units'] + ')  | ')
            # Iter over CSV and write line for line
            with open(self.name + '-chronology.csv', 'r') as f:
                for line in iter(f):
                    line = line.split(',')
                    for index, value in enumerate(line):
                        if index == 0:
                            noaa_txt.write('#          ' + str(value) + '              ')
                        elif index == len(line) - 1:
                            noaa_txt.write(str(value))
                        else:
                            noaa_txt.write(str(value) + '                 ')
        noaa_txt.write('#------------------\n')
        return

    def __write_numbers(self, noaa_txt, table):
        """
        Read numeric data from csv and write to the bottom section of the txt file.
        :param noaa_txt:(obj) Output text file
        :param table: (dict) Paleodata dictionary
        :return: none
        """

        filename = self.__get_filename(table)
        mv = self.__get_mv(table)
        noaa_txt.write('# Data:\
        \n# Data lines follow (have no #)\
        \n# Data line format - tab-delimited text, variable short name as header)\
        \n# Missing_Values: ' + mv + '\n#\n')

        if self.__csv_found(filename):
            # Write variables line from dict_in
            count = len(table['columns'])
            for col in table['columns']:
                try:
                    param = col['parameter']
                except KeyError:
                    param = 'N/A'
                if count == 1:
                    noaa_txt.write("{:<0}".format(param))
                else:
                    noaa_txt.write("{:<15}".format(param))
                    count -= 1
            noaa_txt.write('\n')
            # Iter over CSV and write line for line
            with open(filename, 'r') as f:
                for line in iter(f):
                    line = line.split(',')
                    for index, value in enumerate(line):
                        if index == len(line) - 1:
                            noaa_txt.write("{:<0}".format(str(value)))
                        else:
                            noaa_txt.write("{:<15.10}".format(str(value)))
        noaa_txt.write('\n#\n#------------------\n')
        return

    def __write_variables(self, noaa_txt, table):
        """
        Retrieve variables from data table(s) and write to Variables section of txt file.
        :param noaa_txt:(obj) Output text file
        :param table: (dict) Paleodata dictionary
        :return: none
        """
        filename = self.__get_filename(table)
        noaa_txt.write('# Variables\n#\
            \n# Filename: ' + filename +
            '\n#\n# Data variables follow that are preceded by "##" in columns one and two.\
            \n# Data line variables format:  Variables list, one per line, shortname-tab-longname-tab-longname components ( 9 components: what, material, error, units, seasonality, archive, detail, method, C or N for Character or Numeric data)\n#\n')
        for col in table['columns']:
            # Write one line for each column. One line has all metadata for one column.
            for entry in ORDERING[11]:
                # ---FIX--------FIX--------FIX--------FIX--------FIX-----
                # May need a better way of handling this in the future. Need a strict list for this section.
                try:
                    if entry == 'parameter':
                        # First entry: Double hash and tab
                        noaa_txt.write('{:<20}'.format('##' + col[entry]))
                    elif entry == 'dataType':
                        # Last entry: No space or comma
                        noaa_txt.write('{:<0}'.format(col[entry]))
                    else:
                        # Space and comma after middle entries
                        noaa_txt.write('{:<3}'.format(col[entry] + ', '))
                except KeyError:
                    noaa_txt.write('{:<3}'.format(', '))
            noaa_txt.write('\n')
        noaa_txt.write('#\n#------------------\n')
        return

    def __reorganize_geo(self, d):
        """
        Concat geo value and units, and reorganize the rest
        :param d:
        :return:
        """

        # The new dict that will be returned
        d_tmp = {}

        # Properties
        for k, v in d['properties'].items():
            if k == 'elevation':
                d_tmp['elevation'] = str(v['value']) + ' ' + str(v['unit'])
            else:
                d_tmp[k] = v

        # Geometry
        d_tmp = self.__coordinates(d['geometry']['coordinates'], d_tmp)

        return d_tmp

    def __reorganize(self, key, value):
        """

        :param dict_in:
        :param key:
        :param value:
        :return:
        """

        # If the key isn't in any list, stash it in number 13 for now
        number = 13

        if key in SECTIONS[1]:
            # StudyName only triggers once, append to section 3 also
            if key == 'studyName':
                self.steps_dict[3][key] = value
            number = 1
        elif key in SECTIONS[2]:
            number = 2
        elif key in SECTIONS[4]:
            number = 4
        elif key in SECTIONS[5]:
            number = 5
        elif key in SECTIONS[6]:
            number = 6
        elif key in SECTIONS[7]:
            number = 7
        elif key in SECTIONS[8]:
            number = 8
        elif key in SECTIONS[9]:
            number = 9
        elif key in SECTIONS[10]:
            number = 10
        elif key in SECTIONS[11]:
            number = 11
        elif key in SECTIONS[12]:
            number = 12
        self.steps_dict[number][key] = value

        return

    def __path_context(self, flat_file):
        """
        Turns the flattened json list back in to a usable dictionary structure
        :param flat_file:
        :return:
        """

        new_dict = {}

        # Lists to recompile Values and Units
        elev = []
        core = []

        # Print out each item in the list for debugging
        for item in flat_file:
            split_list = self.__split_path(item)
            lst_len = len(split_list)
            value = split_list[lst_len-1]

            if 'Latitude' in split_list:
                if 'Max' in split_list:
                    new_dict['Northernmost_Latitude'] = value

                elif 'Min' in split_list:
                    new_dict['Southernmost_Latitude'] = value

            elif 'Longitude' in split_list:
                if 'Max' in split_list:
                    new_dict['Easternmost_Longitude'] = value

                elif 'Min' in split_list:
                    new_dict['Westernmost_Longitude'] = value

            elif 'Elevation' in split_list:
                elev.append(value)

            elif 'CoreLength' in split_list:
                core.append(value)

            else:
                if len(split_list) > 2:
                    key = lst_len - 2
                    new_dict[split_list[key]] = value

                else:
                    new_dict[split_list[0]] = split_list[1]

        if core:
            new_dict['Core_Length'] = self.concat_units(core)
        if elev:
            new_dict['Elevation'] = self.concat_units(elev)

        return new_dict



from tkinter import *
from PIL import Image, ImageTk


class MapFrame(Frame):
    """
    Uses Tkinter and PIL to display a map image to the user.
    Source: http://hci574.blogspot.com/2010/04/using-google-maps-static-images.html
    Source: http://www.daniweb.com/forums/thread65288.html
    """
    def __init__(self, master, im, title):
        Frame.__init__(self, master)
        self.caption = Label(self, text=title)
        self.caption.grid()
        self.image = ImageTk.PhotoImage(im) # <--- results of PhotoImage() must be stored
        self.image_label = Label(self, image=self.image, bd=0) # <--- will not work if 'image = ImageTk.PhotoImage(im)'
        self.image_label.grid()
        self.grid()



import zipfile
import shutil


def re_zip(dir_tmp, name, name_ext):
    """
    Zips up directory back to the original location
    :param dir_tmp: (str) Path to tmp directory
    :param name: (str) Name of current file
    :param name_ext: (str) Name of current file with '.lpd' extension
    """
    shutil.make_archive(name_ext, format='zip', root_dir=dir_tmp, base_dir=name)
    return


def unzip(name_ext, dir_tmp):
    """
    Unzip .lpd file contents to tmp directory.
    :param name_ext: (str) Name of lpd file with extension
    :param dir_tmp: (str) Tmp folder to extract contents to
    :return: None
    """

    # Unzip contents to the tmp directory
    try:
        with zipfile.ZipFile(name_ext) as f:
            f.extractall(dir_tmp)
    except FileNotFoundError:
        shutil.rmtree(dir_tmp)
        return
    return
